from .decomposable_score import DecomposableScore
from .gnies_score import GnIESScore
from .interventional import InterventionalScore,FixedInterventionalScore
from .null_score import NullObsScore
from .null_score import NullIntScore



# Copyright 2021 Juan Luis Gamella Martin

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Module containing the code to compute the log-likelihood of a
model (connectivity, noise term means/variances) from data of a single
or multiple environments.
"""

import numpy as np

# --------------------------------------------------------------------
# Functions to compute the full/local log-likelihood


def full(B, noise_means, noise_variances, XX):
    p = len(B)
    likelihood = 0
    for e, X in enumerate(XX):
        inv_omega = np.diag(1 / noise_variances[e])
        inv_cov = (np.eye(p) - B.T).T @ inv_omega @ (np.eye(p) - B.T)
        log_term = -len(X) * np.log(1 / noise_variances[e].prod())
        mu = np.linalg.inv(np.eye(p) - B.T) @ noise_means[e]
        emp_cov = corr(X - mu)
        trace_term = np.trace(emp_cov @ inv_cov)
        # print(log_term, trace_term)
        likelihood += log_term + trace_term
    return -0.5 * likelihood


def local(j, b, noise_means, noise_variances, XX):
    return local_raw(j, b, noise_means, noise_variances, XX)


def full_raw(B, noise_means, noise_variances, XX):
    """Compute the log-likelihood of the given weights (i.e. B), noise
    term means and variances.

    Parameters
    ----------
    B : np.array
        The connectivity matrix.
    noise_means : np.array
        The means of the noise terms of each variable.
    noise_variances : np.array
        The noise term variances of each variable.
    XX : list(np.array)
        The complete joint observations for each environment, where
        each element of the list is an np.array of size n_e x p, where
        p is the total number of variables (same size as b) and n_e is
        the number of observations from that environment.

    Returns
    -------
    likelihood : float
        The log-likelihood of the given parameters

    """
    p = len(B)
    likelihood = 0
    for e, X in enumerate(XX):
        inv_omega = np.diag(1 / noise_variances[e])
        inv_cov = (np.eye(p) - B.T).T @ inv_omega @ (np.eye(p) - B.T)
        log_term = -len(X) * np.log(1 / noise_variances[e].prod())
        # In tests, both ways of computing the log term seemed equally
        # fast, for p=5, n=10K
        # log_term = -len(X) * np.log(np.linalg.det(inv_cov))
        # assert np.isclose(log_term, -len(X) * np.log(np.linalg.det(inv_cov)))
        aux = X - np.linalg.inv(np.eye(p) - B.T) @ noise_means[e]
        cov_term = 0
        for i, x in enumerate(aux):
            cov_term += x @ inv_cov @ x
        likelihood += log_term + cov_term
        # print(log_term, cov_term)
        # In tests, using the trace was almost 4x slower for p=5, n=10K
        # trace_term = np.trace(aux @ inv_cov @ aux.T)
        # likelihood += log_term + trace_term
    return -0.5 * likelihood


def local_raw(j, b, noise_means, noise_variances, XX):
    """Compute the conditional likelihood of the marginal observations of
    a variable given observations of its hypothetical parents, given
    the weights (b) for those parents and the variable's error term
    mean and variance over each environment.

    Parameters
    ----------
    j : int
        The variable's index.
    b : np.array
        A p-sized array with the weights for the parents of j and 0s
        elswhere.
    noise_means : np.array
        A one-dimensional array with the mean of j's noise term for each
        environment.
    noise_variances : np.array
        A one-dimensional array with the variance of j's noise term for each
        environment.
    XX : list(np.array)
        The complete joint observations for each environment, where
        each element of the list is an np.array of size n_e x p, where
        p is the total number of variables (same size as b) and n_e is
        the number of observations from that environment.

    Returns
    -------
    likelihood : float
        The resulting log-likelihood.

    """
    likelihood = 0
    # Iterate over environments
    for e, X in enumerate(XX):
        log_term = len(X) * np.log(noise_variances[e])
        aux = X[:, j] - X @ b - noise_means[e]
        sum_term = aux @ aux / noise_variances[e]
        likelihood += -0.5 * (log_term + sum_term)
    return likelihood


def corr(X, scale=True):
    """Return the empirical correlation matrix given observations from a
    random vector X.

    Parameters
    ----------
    X : array_like
        The observations of the random vector, with columns
        corresponding to variables.
    normalize : bool
        If the resulting matrix should be scaled by the number of
        observations. Defaults to True.

    Returns
    -------
    corr : np.ndarray
        The correlation matrix.

    """
    corr = np.dot(X.T, X) if scale else np.dot(X.T, X) / len(X)
    return corr



# Copyright 2020 Juan Luis Gamella Martin

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""This module contains the implementation of the GnIES score (see
"Score Function" in section 4 of the GnIES paper).

GnIES paper: <TODO: arxiv link>

"""

import numpy as np
import gnies.scores.log_likelihood as log_likelihood
import gnies.scores.log_likelihood_means as log_likelihood_means
from gnies.scores.decomposable_score import DecomposableScore
import gnies.utils as utils

# --------------------------------------------------------------------
# GnIES Score Class

class GnIESScore(DecomposableScore):
    """The class contains a cached implementation of the GnIES score
    presented in the paper.

    Parameters
    ----------
    I : set of ints
        The intervention targets.
    e : int
        The number of environments.
    p : int
        The number of variables.
    n_obs : list of ints
        The number of observations from each environment.
    N : int
        The total number of observations from all environments.
    lmbda : float
        The penalization parameter used in the computation of the score.
    max_iter : float
        The maximum number of iterations for the alternating
        optimization procedure.
    tol : float
        The threshold of convergence for the alternating optimization
        procedure; when the maximum distance between elements of
        consecutive Bs is below this threshold, stop the alternating
        optimization and return the latest estimate.
    centered : bool
        Whether the data is centered before computing the score
        (`centered=True`), or the noise-term means are also estimated
        respecting the constraints imposed by `I`.

    """

    def __init__(self, data, I, centered=True, lmbda=None, tol=1e-16, max_iter=10, cache=True):
        """Creates a new instance of the score.

        Parameters
        ----------
        data : list of numpy.ndarray
            A list with the samples from the different environments, where
            each sample is an array with columns corresponding to
            variables and rows to observations.
        I : set of ints
            The intervention targets.
        centered : bool, default=True
            Whether the data is centered before computing the score
            (`centered=True`), or the noise-term means are also
            estimated respecting the constraints imposed by `I`. The
            latter incurrs a larger penalization term as more
            parameters have to be estimated.
        lmbda : float
            The penalization parameter used in the computation of the score.
        tol : float
            The threshold of convergence for the alternating
            optimization procedure; when the maximum distance between
            estimated coefficients and variances is below this
            threshold, the alternating optimization procedure is
            halted.
        max_iter : float
            The maximum number of iterations for the alternating
            optimization procedure.
        cache : bool, default=True
            Whether the computation of the local score should be cached.

        Raises
        ------
        TypeError :
            If data is not a list of numpy.ndarray or I is not a set.
        ValueError :
            If I is not a subset of {0,...,p-1}; if the given samples
            have different number of variables, or a sample has a
            single observation is given (at least two are needed to
            compute the sample covariances)

        Examples
        --------

        >>> data = list(rng.uniform(size=(5,1000,20)))
        >>> GnIESScore(data, I=set())
        <__main__.GnIESScore object at 0x...>

        >>> bad_data = data.copy()
        >>> bad_data[0] = bad_data[0][:,:-1]
        >>> GnIESScore(bad_data, set())
        Traceback (most recent call last):
        ...
        ValueError: All samples must have the same number of variables.

        >>> bad_data = data.copy()
        >>> bad_data[0] = bad_data[0][[0],:]
        >>> GnIESScore(bad_data, set())
        Traceback (most recent call last):
        ...
        ValueError: Each sample must contain at least two observations to estimate the covariance matrix.

        >>> GnIESScore(data, {20})
        Traceback (most recent call last):
        ...
        ValueError: I must be a subset of {0,...,p-1}.

        """
        # Check input: data
        type_msg = "data should be a list of numpy.ndarray."
        if isinstance(data, list):
            # Check that all samples are arrays
            for i, x in enumerate(data):
                if not isinstance(x, np.ndarray):
                    raise TypeError(type_msg)
            # Check that all samples have the same number of variables
            ps = np.array([x.shape[1] for x in data])
            if len(np.unique(ps)) > 1:
                raise ValueError(
                    "All samples must have the same number of variables.")
            p = data[0].shape[1]
            # Check that all samples have at least two observations
            n_obs = np.array([len(sample) for sample in data])
            if (n_obs == 1).any():
                raise ValueError(
                    "Each sample must contain at least two observations to estimate the covariance matrix.")
        else:
            raise TypeError(type_msg)

        # Check input: I
        if isinstance(I, set):
            if not I <= set(range(p)):
                raise ValueError("I must be a subset of {0,...,p-1}.")
        else:
            raise TypeError("I must be a set of ints.")

        # Intialize parameters
        super().__init__(data, cache=cache)
        self.I = I.copy()
        self.e = len(self._data)
        self.p = self._data[0].shape[1]
        self.n_obs = np.array([len(env) for env in self._data])
        self.N = sum(self.n_obs)
        self.lmbda = 0.5 * np.log(self.N) if lmbda is None else lmbda
        self.max_iter = max_iter
        self.tol = tol
        self.centered = centered

        # Precompute scatter matrices
        self._sample_covariances = np.array([np.cov(env, rowvar=False, ddof=0) for env in data])
        self._pooled_covariance =  np.sum(self._sample_covariances * np.reshape(self.n_obs, (self.e, 1, 1)), axis=0) / self.N

        # Precompute sample means
        if not centered:
            self._sample_means = np.array([np.mean(env, axis=0) for env in self._data])
            self._pooled_means = np.sum(self._sample_means * np.reshape(self.n_obs, (self.e, 1)), axis=0) / self.N

    def set_I(self, new_I):
        """Update the intervention targets used to compute the score, removing
        from the cache all local score computations which are invalidated by
        the change, i.e. those of nodes who were added or removed from I.

        Note: prune_cache is implemented in the parent class
        `DecomposableScore`.

        Parameters
        ----------
        new_I : set of ints
            The new intervention targets.

        Raises
        ------
        ValueError :
            If new_I is not a subset of {0,...,p-1}.
        """
        if not new_I <= set(range(self.p)):
            raise ValueError("I must be a subset of {0,...,p-1}.")

        change = (self.I - new_I) | (new_I - self.I)
        self.prune_cache(change)
        self.I = new_I.copy()

    def full_score(self, A):
        """
        Given a DAG adjacency A, return its score by finding
        the maximum likelihood estimates of the corresponding connectivity
        matrix (weights) and noise term variances.

        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j

        Returns
        -------
        score : float
            the penalized log-likelihood score

        Raises
        ------
        ValueError :
            If the given adjacency is not a DAG or is not a `self.p x
            self.p` matrix.
        """
        # Compute MLE and its likelihood
        B, omegas, means = self._mle_full(A)
        if self.centered:
            likelihood = log_likelihood.full(B, omegas, self._sample_covariances, self.n_obs)
        else:
            likelihood = log_likelihood_means.full(B, means, omegas, self._data)
        # Penalization term
        l0_term = self.lmbda * self._ddof_full(A)
        score = likelihood - l0_term
        return score

    def local_score(self, j, pa):
        """Given a node and its parents, compute its associated term in the score.

        Note: the computation is implemented in
        `_compute_local_score`; local_score in the parent class
        `DecomposableScore` implements the cached wrapper.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : set of ints
            The parents of the variable

        Returns
        -------
        score : float
            The local score.

        """
        return super().local_score(j, pa)

    def _compute_local_score(self, j, pa):
        """
        Given a node and its parents, compute its associated term in the score.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : set of ints
            The parents of the variable

        Returns
        -------
        score : float
            The local score.

        """
        b, omegas, means = self._mle_local(j, pa)
        # Compute likelihood
        if self.centered:
            likelihood = -0.5 * ((1 + np.log(omegas)) * self.n_obs).sum()
        else:
            likelihood = log_likelihood_means.local(j, b, means, omegas, self._data)
        # Penalization term
        l0_term = self.lmbda * self._ddof_local(j, pa)
        score = likelihood - l0_term
        return score

    # --------------------------------------------------------------------
    #  Functions for the maximum likelihood estimation of the
    #  weights/variances
    def _mle_full(self, A):
        """Given the DAG adjacency A, compute the maximum
        likelihood estimate of the connectivity weights and noise-term
        variances and means.

        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j

        Returns
        -------
        B : np.array
            The MLE estimate of the connectivity matrix (the weights).
        omegas : np.array
            The MLE estimate of the noise-term variances for each
            environment and variable, in a `self.e x self.p` array.
            variables.
        means : np.array or NoneType
            The MLE estimate of the noise-term means for each
            environment and variable, in a `self.e x self.p` array. Is
            `None` if `self.centered=True`.

        Raises
        ------
        ValueError :
            If the given adjacency is not a DAG or is not a `self.p x
            self.p` matrix.

        """
        # Check inputs
        if A.shape != (self.p, self.p):
            raise ValueError('A is not a %d x %d matrix' % (self.p, self.p))
        elif not utils.is_dag(A):
            raise ValueError('The given adjacency A is not a DAG.')
        # Compute score
        B = np.zeros((self.p, self.p), dtype=float)
        Omegas = np.zeros((self.e, self.p), dtype=float)
        Means = []
        for j in range(self.p):
            pa = utils.pa(j, A)
            b, omegas, means = self._mle_local(j, pa)
            B[:, j], Omegas[:, j] = b, omegas
            Means.append(means)
        Means = None if self.centered else np.array(Means).T
        return B, Omegas, Means

    def _mle_local(self, j, pa):
        """Given a variable and its parents, computes the MLE of the
        regression coefficients and noise-term variances and means.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : set of ints
            The parents of the variable

        Returns
        -------
        b : np.array
            The MLE of the regression coefficients, with length as
            many elements in `pa`.
        omegas : np.array
            The MLE of the noise-term variances, in an array with
            length the number of environments `self.e`.
        means : np.array or NoneType
            The MLE of the noise-term means, in an array with length
            the number of environments `self.e`. Is `None` if
            `self.centered` is True.
        """
        pa = sorted(pa)
        b, omegas = self._alternating_mle(j, pa, debug=0)
        means = None if self.centered else self._noise_means_from_b(j, pa, b)
        b = _embedd(b, self.p, pa)
        return b, omegas, means

    def _omegas_from_b(self, j, pa, b):
        """Given its parents and corresponding regression coefficients,
        estimate the noise-term variances of the given variable.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : list of ints
            The parents of the variable
        b : np.array
            The regression coefficients, with length as many elements
            in `pa`.

        Returns
        -------
        omegas : np.array
            The estimate of noise-term variances of the variable, in
            an array of length `self.e`.

        """
        # variable j has not received interventions: its noise-term
        # variance is constant across environments
        if j not in self.I:
            omega = self._pooled_covariance[j, j] - self._pooled_covariance[j, pa] @ b
            omegas = np.ones(self.e, dtype=float) * omega
        if j in self.I:
            I_B = -_embedd(b, self.p, pa)
            I_B[j] = 1
            omegas = I_B @ self._sample_covariances @ I_B
            # for e,cov in enumerate(sample_covariances):
            #     omegas[e] = cov[j,j] - cov[j,pa] @ b
            #     # Why does the above (commented) not work?  Idea: We're
            #     # not asking about the variance of the conditional
            #     # distribution of Xj given its parents in environment e,
            #     # as this would mean the regression coefficients would be
            #     # allowed to change within environments. We want to know
            #     # the variance of the residuals resulting from using the
            #     # regression coefficients in B.
        return omegas

    def _noise_means_from_b(self, j, pa, b):
        """Given its parents and corresponding regression coefficients,
        estimate the noise-term means of the given variable.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : list of ints
            The parents of the variable
        b : np.array
            The regression coefficients, with length as many elements
            in `pa`.

        Returns
        -------
        means : np.array
            The estimate of noise-term means of the variable, in
            an array of length `self.e`.

        """
        if j not in self.I:
            mean = self._pooled_means[j] - self._pooled_means[pa] @ b
            means = np.ones(self.e, dtype=float) * mean
        else:
            means = self._sample_means[:, j] - self._sample_means[:, pa] @ b
        return means

    def _b_from_omegas(self, j, pa, omegas):
        """Given its noise-term variances across environments and parents,
        estimate the regression coefficients for the given variable.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : list of ints
            The parents of the variable
        omegas : np.array
            The noise-term variances of the variable, in an array of
            length `self.e`.

        Returns
        -------
        b : np.array
            The estimate of the regression coefficients, with length as
            many elements in `pa`.

        """
        if j not in self.I:
            weighted_covariance = self._pooled_covariance
        else:
            weights = self.n_obs / omegas
            weights /= sum(weights)
            weighted_covariance = np.sum(self._sample_covariances * np.reshape(weights, (self.e, 1, 1)), axis=0)
        return _regress(j, pa, weighted_covariance)

    def _alternating_mle(self, j, pa, debug=False):
        """The alternating optimization procedure used to estimate the
        regression coefficients and noise term variances of a variable
        given its parents.

        Parameters
        ----------
        j : int
            The index of the node.
        pa : list of ints
            The parents of the variable
        debug : bool, default=False
            If debugging traces should be printed.

        Returns
        -------
        b : np.array
            The MLE of the regression coefficients, with length as
            many elements in `pa`.
        omegas : np.array
            The MLE of the noise-term variances, in an array with
            length the number of environments `self.e`.
        """
        # If j has no parents no alternating is needed
        if len(pa) == 0:
            b = np.array([])
            omegas = self._omegas_from_b(j, pa, b)
            return b, omegas
        else:
            # Set starting point for optimization procedure
            prev_b = b = _regress(j, pa, self._pooled_covariance)
            prev_omegas = omegas = self._omegas_from_b(j, pa, b)
            # Start alternating optimization
            for i in range(self.max_iter):
                omegas = self._omegas_from_b(j, pa, b)
                b = self._b_from_omegas(j, pa, omegas)
                delta = max(abs(b - prev_b).max(), abs(omegas - prev_omegas).max())
                if debug:
                    print(" %0.4f" % delta, end="")
                # Check stopping conditions
                if delta <= self.tol:
                    print() if debug else None
                    return b, omegas
                else:
                    prev_omegas = omegas
                    prev_b = b
            print(" MAX ITER REACHED") if debug else None
            return b, omegas

    def _ddof_full(self, A):
        """Compute the number of free parameters in a model specified by the
        DAG adjacency and the intervention targets.

        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j.

        Returns
        -------
        ddof : int
            the number of free parameters in the model


        """
        n_edges = (A != 0).sum()
        n_variances = len(self.I) * self.e + (self.p - len(self.I))
        n_intercepts = 0 if self.centered else n_variances
        return n_edges + n_variances + n_intercepts


    def _ddof_local(self, j, pa):
        """Compute the number of free parameters that are estimated for a
        single variable in the model, given its parents and the
        intervention targets.

        Parameters
        ----------
        j : int
            The variable's index.
        pa : list(int)
            The set of parents.

        Returns
        -------
        ddof : int
            the number of free parameters

        """
        n_variances = self.e if j in self.I else 1
        n_intercepts = 0 if self.centered else n_variances
        return len(pa) + n_variances + n_intercepts


# --------------------------------------------------------------------
# Support functions for the alternating optimization procedure

def _regress(j, pa, cov):
    """Compute the regression coefficients from the covariance matrix i.e.
    b = Σ_{j,pa(j)} @ Σ_{pa(j), pa(j)}^-1
    """
    return np.linalg.solve(cov[pa, :][:, pa], cov[j, pa])


def _embedd(b, p, idx):
    """Create a new vector with elements in idx corresponding to the
    elements in b and zeros elsewhere."""
    vector = np.zeros(p, dtype=b.dtype)
    vector[idx] = b
    return vector

# --------------------------------------------------------------------
# Discarded implementation of the alternating optimization procedure
# (equivalent to the current one)

# def _alternating_mle(self, j, pa, debug=0):
#         """Estimate the parameters for variable j, i.e. the parent weights and
#         variances (means) of the noise terms."""
#         # If j has no parents, can directly estimate omegas
#         if len(pa) == 0:
#             b = np.array([])
#             omegas = self._omegas_from_b(j, pa, b)
#             return b, omegas
#         else:
#             # Set starting point for optimization procedure
#             b_0 = _regress(j, pa, self._pooled_covariance)
#             omegas_0 = self._omegas_from_b(j, pa, b_0)

#             # Components for the alternating (EM-like) optimization procedure
#             # "Expectation" function: given b compute the noise-term variances
#             def e_func(b):
#                 return self._omegas_from_b(j, pa, b)

#             # "Maximization" function: given variances of noise terms, compute B
#             def m_func(omegas):
#                 return self._b_from_omegas(j, pa, omegas)

#             # Distance to check for convergence: max of L1 distance between
#             # successive b's and successive omegas
#             def dist(prev_b, prev_omegas, b, omegas):
#                 return max(abs(b - prev_b).max(), abs(omegas - prev_omegas).max())

#             # Keep track of the objective function if debugging is desired
#             if debug:
#                 def objective(B, omegas):
#                     return log_likelihood.local(j, b, omegas, self._sample_covariances, self.n_obs)
#             else:
#                 objective = None

#             # Run procedure
#             print("    Running alternating optimization procedure") if debug else None
#             (b, omegas) = _em_like(e_func, m_func, b_0, omegas_0, dist,
#                                    tol=self.tol, max_iter=self.max_iter, objective=objective, debug=debug)
#             return b, omegas

# def _em_like(e_func, m_func, M_0, E_0, dist, tol=1e-6, assume_convex=False, max_iter=100, objective=None, debug=False):
#     """
#     EM-like alternating optimization procedure.

#     Parameters
#     ----------
#     e_func : function
#         "Expectation" function
#     m_func : function
#         "Maximization" function
#     M_0 : any
#         initial value for the expectation function
#     E_0 : any
#         initial value for the maximization function. In principle, not
#         needed as the first step of the algorithm overrides this
#         value; however, if the initial values satisfy convergence this
#         allows us to return them : in a user-transparent way).
#     dist : function
#         The distance function used to check convergence. Takes as
#         arguments the E,M from the previous and current iterations and
#         returns a float
#     tol : float
#         Convergence tolerance. Stop if distance between consecutive
#         iterations is below this threshold.
#     assume_convex : boolean
#         if the objective is assumed convex. If true, an increase in
#         the distance of iterations is due to numerical issues, and we
#         stop the iterations at this point, even if the tolerance has
#         not been reached
#     objective : function
#         For debugging purposes, the objective function
#     debug : boolean
#         Print results of each iteration

#     Returns
#     -------
#     M : any
#         the approximated maximizer of the "Maximization quantity"
#     E : any
#         the approximated maximizer of the "Expectation quantity"

#     """
#     print("     ", end="") if debug else None
#     prev_E = E = E_0
#     prev_M = M = M_0
#     prev_delta = np.inf
#     for i in range(max_iter):
#         E = e_func(M)  # E-step
#         M = m_func(E)  # M-step
#         delta = dist(prev_M, prev_E, M, E)
#         # Debug outputs
#         if objective is not None and debug:
#             try:
#                 value = objective(M, E)
#                 print(" %0.16f (%0.16f)" % (delta, value), end="")
#             except Exception as e:
#                 print(" %0.16f (%s)" % (delta, e), end="")
#         elif debug:
#             print(" %0.4f" % delta, end="")
#             # Check stopping conditions
#         if delta <= tol:
#             print() if debug else None
#             return M, E
#         elif assume_convex and delta > prev_delta:
#             # if program is convex, an increase in the distance
#             # between iterations is due to numerical issues. Thus we
#             # return the results from the previous iteration as we
#             # assume that's as close as we can get
#             print() if debug else None
#             return prev_M, prev_E
#         else:
#             prev_E = E
#             prev_M = M
#             prev_delta = delta
#     print(" MAX ITER REACHED") if debug else None
#     return M, E

if __name__ == '__main__':
    import doctest
    doctest.testmod(extraglobs={'rng': np.random.default_rng(42)},
                    verbose=True,
                    optionflags=doctest.ELLIPSIS)



# Copyright 2020 Juan Luis Gamella Martin

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
This module contains classes implementing interventional scores with different definitions
for the intervention targets.
"""

import numpy as np
from . import log_likelihood
from . import log_likelihood_means
from .decomposable_score import DecomposableScore

# --------------------------------------------------------------------
# Interventional Score Class (see FixedInterventionalScore below for GnIES score)

class InterventionalScore(DecomposableScore):
    """Class implementing interventional scores with different definitions
    for the intervention targets: per environment or pooled (see
    `fine_grained` parameter).

    The GnIES score is the score with `fine_grained=False` and is
    implemented in the `FixedInterventionalScore` class, which
    inherits from this one.

    Parameters
    ----------
    e : int
        The number of environments.
    p : int
        The number of variables.
    n_obs : list of ints
        The number of observations from each environment.
    N : int
        The total number of observations from all environments.
    lmbda : float
        The penalization parameter used in the computation of the score.
    max_iter : float
        The maximum number of iterations for the alternating
        optimization procedure.
    tol : float
        The threshold of convergence for the alternating optimization
        procedure; when the maximum distance between elements of
        consecutive Bs is below this threshold, stop the alternating
        optimization and return the latest estimate.
    fine_grained : bool
        How finely we specify intervention targets. `True` means we
        define intervention targets per environment; `False` means we
        allow the noise-term distribution of an intervened variable to
        vary across all environments. The latter is used for the
        computation of the GnIES score.
    centered : bool
        Whether the data is centered before computing the score
        (`centered=True`), or the noise-term means are also estimated
        respecting the constraints imposed by `I`.

    """

    def __init__(self, data, centered=True, fine_grained=True, lmbda=None, tol=1e-16, max_iter=10):
        """
        Creates a new instance of the score.

        Parameters
        ----------
        data : list of numpy.ndarray
            A list with the samples from the different environments, where
            each sample is an array with columns corresponding to
            variables and rows to observations.
        centered : bool
            Whether the data is centered before computing the score
            (`centered=True`), or the noise-term means are also estimated
            respecting the constraints imposed by `I`.
        fine_grained : bool
            How finely we specify intervention targets. `True` means we
            define intervention targets per environment; `False` means we
            allow the noise-term distribution of an intervened variable to
            vary across all environments. The latter is used for the
            computation of the GnIES score.
        lmbda : float
            The penalization parameter used in the computation of the score.
        tol : float
            The threshold of convergence for the alternating optimization
            procedure; when the maximum distance between elements of
            consecutive Bs is below this threshold, stop the alternating
            optimization and return the latest estimate.
        max_iter : float
            The maximum number of iterations for the alternating
            optimization procedure.
        """
        super().__init__(data, cache=False)
        self.e = len(data)
        self.p = data[0].shape[1]
        self.n_obs = np.array([len(env) for env in data])
        self.N = sum(self.n_obs)
        self.lmbda = 0.5 * np.log(self.N) if lmbda is None else lmbda
        self.max_iter = max_iter
        self.tol = tol
        self.fine_grained = fine_grained
        self.centered = centered

        # Precompute scatter matrices
        self._sample_covariances = np.array([np.cov(env, rowvar=False, ddof=0) for env in data])
        # Precompute sample means
        if not centered:
            self._sample_means = np.array([np.mean(env, axis=0) for env in data])

    def full_score(self, A, I):
        """
        Given a DAG adjacency A, return the l0-penalized log-likelihood of
        a collection of samples from different environments, by finding
        the maximum likelihood estimates of the corresponding connectivity
        matrix (weights) and noise term variances.

        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j
        I : list of sets
            the sets of variables which have received interventions in
            each environment

        Returns
        -------
        score : float
            the penalized log-likelihood score

        """
        I = I if self.fine_grained else [I] * self.e
        # Compute MLE and its likelihood
        if self.centered:
            B, omegas = self._mle_full(A, I)
            likelihood = log_likelihood.full(B, omegas, self._sample_covariances, self.n_obs)
        else:
            B, nus, omegas = self._mle_full(A, I)
            likelihood = log_likelihood_means.full(B, nus, omegas, self._data)
        #   Note: the number of parameters is the number of edges +
        #   the total number of marginal variances/means, which depends on
        #   the number of interventions.
        l0_term = self.lmbda * ddof_full(A, I, centered=self.centered)
        score = likelihood - l0_term
        return score

    def local_score(self, x, pa, I):
        """Return the local score of a given node and a set of
        parents. If self.cache=True, will use previously computed
        score if possible.

        Parameters
        ----------
        x : int
            a node
        pa : set of ints
            the node's parents
        I : list of sets
            the sets of variables which have received interventions in
            each environment

        Returns
        -------
        score : float
            the corresponding score

        """
        I = I if self.fine_grained else [I] * self.e
        return self._compute_local_score(x, pa, I)
        # if self._cache is None:
        #     return self._compute_local_score(x, pa, I)
        # else:
        #     key = (x, tuple(sorted(pa)))
        #     try:
        #         score = self._cache[key]
        #         print("score%s: using cached value %0.2f" % (key,score)) if self._debug>=2 else None
        #     except KeyError:
        #         score = self._compute_local_score(x, pa, I)
        #         self._cache[key] = score
        #         print("score%s = %0.2f" % (key,score)) if self._debug>=2 else None
        #     return score

    def _compute_local_score(self, x, pa, I):
        """
        Given a node and its parents, return the local l0-penalized
        log-likelihood of a collection of samples from different
        environments, by finding the maximum likelihood estimates of the
        weights and noise term variances.

        Parameters
        ----------
        x : int
            a node
        pa : set of ints
            the node's parents
        I : list of sets
            the sets of variables which have received interventions in
            each environment

        Returns
        -------
        score : float
            the penalized log-likelihood score

        """
        # Compute MLE, with the local subgraph p -> x for p in pa
        #   Only do set -> list conversion once, as the ordering is not
        #   guaranteed to be consistent.
        if self.centered:
            b, omegas = self._mle_local(x, pa, I)
            likelihood = log_likelihood.local(x, b, omegas, self._sample_covariances, self.n_obs)
        else:
            b, nus, omegas = self._mle_local(x, pa, I)
            likelihood = log_likelihood_means.local(x, b, nus, omegas, self._data)
        #  Note: the number of parameters is the number of parents (one
        #  weight for each) + one marginal variance/mean per environment for x
        l0_term = self.lmbda * ddof_local(x, pa, I, centered=self.centered)
        score = likelihood - l0_term
        return score

    # --------------------------------------------------------------------
    #  Functions for the maximum likelihood estimation of the
    #  weights/variances
    def _mle_full(self, A, I):
        """Given the DAG adjacency A and observations, compute the maximum
        likelihood estimate of the connectivity weights and noise
        variances, returning them and the resulting log likelihood.

        Parameters
        ----------
        A : np.array
            The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j

        Returns
        -------
        B : np.array
            the MLE estimate of the connectivity matrix (the weights)
        omegas : np.array
            the MLE estimate of the noise variances of each variable

        """
        B, omegas = _alternating_mle(A, I, self._sample_covariances,
                                     self.n_obs, self.tol, self.max_iter, self._debug)
        if self.centered:
            return B, omegas
        else:
            noise_term_means = _noise_means_from_B(B, I, self._sample_means, self.n_obs)
            return B, noise_term_means, omegas

    def _mle_local(self, j, pa, I):
        pa = list(pa)
        A = np.zeros((1 + len(pa), 1 + len(pa)))
        A[1:, 0] = 1
        sub_I = [set(range(1, len(pa) + 1)) for _ in I]
        for k, targets in enumerate(I):
            if j in targets:
                # print(k, j, targets)
                sub_I[k].add(0)
        # print(I)
        # print(sub_I)
        sub_covariances = np.array([sigma[[j] + pa, :][:, [j] + pa]
                                    for sigma in self._sample_covariances])
        B, Omegas = _alternating_mle(A, sub_I, sub_covariances, self.n_obs,
                                     self.tol, self.max_iter, self._debug)
        b = np.zeros(self.p)
        b[pa] = B[1:, 0]
        omegas = Omegas[:, 0]  # Extract marginal variances of x
        if self.centered:
            return b, omegas
        else:
            sub_means = self._sample_means[:, [j] + pa]
            nus = _noise_means_from_B(B, sub_I, sub_means, self.n_obs)[:, 0]
            return b, nus, omegas

# --------------------------------------------------------------------
# GnIES Score Class

class FixedInterventionalScore(InterventionalScore):

    def __init__(self, data, I, centered=True, fine_grained=False, lmbda=None, tol=1e-16, max_iter=10):
        self.I = I
        super().__init__(data, centered, fine_grained, lmbda, tol, max_iter)

    def full_score(self, A):
        return super().full_score(A, self.I)

    def local_score(self, x, pa):
        return super().local_score(x, pa, self.I)


# --------------------------------------------------------------------
# Functions for the alternating maximization procedure used to find
# the MLE

def _alternating_mle(A, I, sample_covariances, n_obs, tol=1e-16, max_iter=50, debug=0):
    """Given the DAG adjacency A and observations, compute the maximum
    likelihood estimate of the connectivity weights and noise
    variances using an alternating maximization procedure.

    Parameters
    ----------
    A : np.array
        The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j
    sample_covariances : list(np.array())
        the list of sample covariances (see np.cov) of the sample from
        each environment
    n_obs : list of ints
       the number of observations available from each environment
       (i.e. the sample size)
    tol : float
        The tolerance for the alternating maximization procedure (see
        em_like below) used to find the MLE, i.e. the procedure will
        stop when the maximum L1 distance between succesive parameters
        (weights and variances) is less than tol.
    max_iter : int
        The maximum number of iterations for which the alternating
        maximization procedure is run (see em_like below).
    debug : bool
        if debug traces should be printed

    Returns
    -------
    B : np.array
        the MLE estimate of the connectivity matrix (the weights)
    omegas : np.array
        the MLE estimate of the noise variances of each variable

    """
    # Set starting point for optimization procedure
    p = len(A)
    B_0 = np.zeros((p, p))
    avg_covariance = sample_covariances.mean(axis=0)
    for j in range(p):
        pa = list(np.where(A[:, j] != 0)[0])
        B_0[pa, j] = _regress(j, pa, avg_covariance)
    # Set up alternating minimization procedure:
    omegas_0 = _omegas_from_B(B_0, [set()] * len(I), sample_covariances, n_obs)  # e_func(B_0)

    # 1. "Expectation" function: given B compute variances of noise terms

    def e_func(B):
        return _omegas_from_B(B, I, sample_covariances, n_obs)

    # 2. "Maximization" function: given variances of noise terms, compute B

    def m_func(omegas):
        return _B_from_omegas(omegas, A, sample_covariances, n_obs)
    # 3. Distance to check for convergence: max of L1 distance between B's and between omegas

    def dist(prev_B, prev_omegas, B, omegas):
        return max(abs(B - prev_B).max(), abs(omegas - prev_omegas).max())

    # 4. Keep track of the objective function if debugging is desired

    def objective(B, omegas):
        return log_likelihood.full(B, omegas, sample_covariances, n_obs) if debug else None

    # Run procedure
    print("    Running alternating optimization procedure") if debug else None
    (B, omegas) = _em_like(e_func, m_func, B_0, omegas_0, dist,
                           tol=tol, max_iter=max_iter, objective=objective, debug=debug)
    return B, omegas


def _regress(j, pa, cov):
    # compute the regression coefficients from the
    # empirical covariance (scatter) matrix i.e. b =
    # Σ_{j,pa(j)} @ Σ_{pa(j), pa(j)}^-1
    return np.linalg.solve(cov[pa, :][:, pa], cov[j, pa])


def _B_from_omegas(omegas, A, sample_covariances, n_obs):
    """
    Part of the alternating likelihood maximization procedure. Given
    fixed noise variances and a DAG adjacency A, return the maximizing
    connectivity matrix.

    Parameters
    ----------
    omegas : np.array
        the noise variances of each variable
    A : np.array
        The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j
    sample_covariances : list(np.array())
        the list of sample covariances (see np.cov) of the sample from
        each environment
    n_obs : list of ints
       the number of observations available from each environment
       (i.e. the sample size)

    Returns
    -------
    B : np.array
        the maximizing connectivity matrix

    """
    e = len(sample_covariances)
    p = len(sample_covariances[0])
    B = np.zeros((p, p))
    for j in range(p):
        # Dividing by j's variances seems a bit weird at first, but
        # note that the regression below is done only for j
        weights = (n_obs / omegas[:, j])
        weights /= sum(weights)
        pooled_covariance = np.sum(sample_covariances * np.reshape(weights, (e, 1, 1)), axis=0)
        parents = np.where(A[:, j] != 0)[0]
        coef = _regress(j, parents, pooled_covariance)
        B[parents, j] = coef
    return B


def _noise_means_from_B(B, I, sample_means, n_obs):
    e, p = sample_means.shape
    I_B = (np.eye(p) - B)
    # This gives the noise term means for each environment (as if all
    # variables were intervened, on all environments)
    means = sample_means @ I_B
    # Compute the noise means for each variable across the
    # non-intervened environments
    scaled = sample_means * np.reshape(n_obs, (e, 1))
    for j in range(p):
        non_intervened = np.where([j not in i for i in I])[0]
        if len(non_intervened) > 0:
            # pooled is a 1xp array, containing the mean for each
            # variable, pooled across the "non_intervened" environments
            pooled = scaled[non_intervened].sum(axis=0) / n_obs[non_intervened].sum()
            means[non_intervened, j] = pooled @ I_B[:, j]
    return means


def _omegas_from_B(B, I, sample_covariances, n_obs):
    """
    Part of the alternating likelihood maximization procedure. Given
    fixed noise variances and a DAG adjacency A, return the maximizing
    connectivity matrix.

    Parameters
    ----------
    omegas : np.array
        the noise variances of each variable
    A : np.array
        The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j
    sample_covariances : list(np.array())
        the list of sample covariances (see np.cov) of the sample from
        each environment
    n_obs : list of ints
       the number of observations available from each environment
       (i.e. the sample size)

    Returns
    -------
    B : np.array
        the maximizing connectivity matrix

    """
    e = len(sample_covariances)
    if len(I) != e:
        raise ValueError(
            "Wrong number of intervention targets (%d) for %d environments" % (len(I), e))
    p = len(B)
    omegas = np.zeros((e, p))
    # Scale sample covariances wrt. each environment's sample size
    scaled = sample_covariances * np.reshape(n_obs, (e, 1, 1))
    for j in range(p):
        # parents = np.where(B[:, j] != 0)[0]
        # Separate into environments where j is intervened/not
        intervened = np.where([j in i for i in I])[0]
        non_intervened = np.where([j not in i for i in I])[0]
        # Compute variance for environments where variable was not
        # intervened (variance is fixed across them)
        if len(non_intervened) > 0:
            pooled = scaled[non_intervened].sum(axis=0) / sum(n_obs[non_intervened])
            # variance = pooled[j,j] - pooled[j,parents] @ B[parents,j]
            variance = ((np.eye(p) - B).T @ pooled @ (np.eye(p) - B))[j, j]
            omegas[non_intervened, j] = variance
        # Compute variance for other environments
        for k in intervened:
            # TODO: Fix this computational waste (computing a matrix product to extract one element)
            variance = ((np.eye(p) - B).T @ sample_covariances[k] @ (np.eye(p) - B))[j, j]
            # variance = sample_covariances[k,j,j] - sample_covariances[k,j,parents] @ B[parents,j]
            # Why does the above (commented) not work?  Idea: We're
            # not asking about the variance of the conditional
            # distribution of Xj given its parents in environment e,
            # as this would mean the regression coefficients would be
            # allowed to change within environments. We want to know
            # the variance of the residuals resulting from using the
            # regression coefficients in B.
            omegas[k, j] = variance
    return omegas  # abs(omegas)


def _em_like(e_func, m_func, M_0, E_0, dist, tol=1e-6, assume_convex=False, max_iter=100, objective=None, debug=False):
    """
    EM-like alternating optimization procedure.

    Parameters
    ----------
    e_func : function
        "Expectation" function
    m_func : function
        "Maximization" function
    M_0 : any
        initial value for the expectation function
    E_0 : any
        initial value for the maximization function. In principle, not
        needed as the first step of the algorithm overrides this
        value; however, if the initial values satisfy convergence this
        allows us to return them : in a user-transparent way).
    dist : function
        The distance function used to check convergence. Takes as
        arguments the E,M from the previous and current iterations and
        returns a float
    tol : float
        Convergence tolerance. Stop if distance between consecutive
        iterations is below this threshold.
    assume_convex : boolean
        if the objective is assumed convex. If true, an increase in
        the distance of iterations is due to numerical issues, and we
        stop the iterations at this point, even if the tolerance has
        not been reached
    objective : function
        For debugging purposes, the objective function
    debug : boolean
        Print results of each iteration

    Returns
    -------
    M : any
        the approximated maximizer of the "Maximization quantity"
    E : any
        the approximated maximizer of the "Expectation quantity"

    """
    print("     ", end="") if debug else None
    prev_E = E = E_0
    prev_M = M = M_0
    prev_delta = np.inf
    for i in range(max_iter):
        E = e_func(M)  # E-step
        M = m_func(E)  # M-step
        delta = dist(prev_M, prev_E, M, E)
        # Debug outputs
        if objective is not None and debug:
            try:
                value = objective(M, E)
                print(" %0.16f (%0.16f)" % (delta, value), end="")
            except Exception as e:
                print(" %0.16f (%s)" % (delta, e), end="")
        elif debug:
            print(" %0.4f" % delta, end="")
            # Check stopping conditions
        if delta <= tol:
            print() if debug else None
            return M, E
        elif assume_convex and delta > prev_delta:
            # if program is convex, an increase in the distance
            # between iterations is due to numerical issues. Thus we
            # return the results from the previous iteration as we
            # assume that's as close as we can get
            print() if debug else None
            return prev_M, prev_E
        else:
            prev_E = E
            prev_M = M
            prev_delta = delta
    print(" MAX ITER REACHED") if debug else None
    return M, E


def ddof_full(A, I, centered=True):
    """Compute the number of free parameters in a model specified by the
    DAG adjacency and the intervention targets.

    Parameters
    ----------
    A : np.array
        The adjacency matrix of a DAG, where A[i,j] != 0 => i -> j.
    I : list of sets
        The sets of variables which have received interventions in
        each environment.
    centered : bool
        If we are also fitting an intercept or we assume the data to
        be centered.

    Returns
    -------
    ddof : int
        the number of free parameters in the model


    """
    p = len(A)
    # We estimate a weight for each edge
    no_edges = (A != 0).sum()
    # For each variable, we have to estimate its noise term variance
    # for each environment in which it receives an intervention, and
    # one for the rest (if any)
    no_variances = sum(len(i) for i in I) + p - len(set.intersection(*I))
    # Same goes for the noise term mean
    no_intercepts = 0 if centered else no_variances
    return no_edges + no_variances + no_intercepts


def ddof_local(j, pa, I, centered=True):
    """Compute the number of free parameters that are estimated for a
    variable in the model, given its parents and the intervention
    targets.

    Parameters
    ----------
    j : int
        The variable's index.
    pa : list(int)
        The set of parents.
    I : list of sets
        The sets of variables which have received interventions in
        each environment.
    centered : bool
        If we are also fitting an intercept or we assume the data to
        be centered.

    Returns
    -------
    ddof : int
        the number of free parameters

    """
    # We have to estimate the variance of the noise term for each
    # environment in which j receives an intervention, and one for the
    # rest (if any)
    no_variances = sum(j in i for i in I) + min(1, sum(j not in i for i in I))
    # Same goes for the noise term mean
    no_intercepts = 0 if centered else no_variances
    # We estimate one weight per parent
    return len(pa) + no_variances + no_intercepts



import setuptools

setuptools.setup(
    name="gnies",
    version="0.3.3",
    author="Juan L. Gamella",
    author_email="juangamella@gmail.com",
    packages=["gnies", "gnies.scores"],
    scripts=[],
    url="https://github.com/juangamella/gnies",
    license="BSD 3-Clause License",
    description="",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=["numpy>=1.17.0", "ges>=1.1.0"],
)



# Copyright 2020 Juan Luis Gamella Martin

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Module containing the code to compute the log-likelihood of a model
(connectivity and variances) from data of a single or multiple
environments. This module is different to log_likelihood_means in that
the means are always set to their MLE. This offers the possibility of
using precomputed sample covariance matrices (see full and local
functions).
"""

import numpy as np

# --------------------------------------------------------------------
# Functions to compute the full/local log-likelihood


def full(B, omegas, sample_covariances, n_obs):
    """
    Compute the log-likelihood of the given weights (B) and noise
    variances (omegas).

    Parameters
    ----------
    B : np.array
        the connectivity matrix
    omegas : np.array
        the noise variances of each variable
    sample_covariances : list(np.array())
        the list of sample covariances (see np.cov) of the sample from
        each environment
    n_obs : list of ints
       the number of observations available from each environment
       (i.e. the sample size)

    Returns
    -------
    likelihood : float
        the log-likelihood of the given parameters

    """
    p = len(B)
    likelihood = 0
    for j, sigma in enumerate(sample_covariances):
        K = np.diag(1 / omegas[j])
        det_K = np.prod(1 / omegas[j])
        likelihood += n_obs[j] * (np.log(det_K) - np.trace(K @
                                                           (np.eye(p) - B.T) @ sigma @ (np.eye(p) - B.T).T))
    return likelihood * 0.5


def local(j, b, omegas, sample_covariances, n_obs):
    """
    Compute the conditional likelihood of the marginal observations of
    a variable given observations of its hypothetical parents, given
    the weights (b) for those parents and the variable's error term
    variances over each environment (omegas).

    Parameters
    ----------
    j : int
        the variable's index
    b : np.array
        p-sized array with the weights for the parents of j and 0s
        elswhere
    omegas : np.array
        1d array with the noise term variance of j for each
        environment
    sample_covariances : list(np.array())
        the list of sample covariances (see np.cov) of the sample from
        each environment
    n_obs : list of ints
       the number of observations available from each environment
       (i.e. the sample size)

    Returns
    -------
    log_likelihood : float
        the resulting log_likelihood

    """
    return -0.5 * ((1 + np.log(omegas)) * n_obs).sum()

def full_raw(B, omegas, data):
    """
    Compute the log likelihood of the given connectivity matrix (B) and noise
    variances (omegas).

    Parameters
    ----------
    B : np.array
        the connectivity matrix, where B[i,j] = w iff i -> j with weight w
    omegas : np.array
        the noise variances of each variable
    data : list(np.array)
        A list of samples from each environment, where each sample is
        an np.array with each column corresponding to a variable

    Returns
    -------
    likelihood : float
        the log-likelihood of the given parameters

    """
    sample_covariances = [np.cov(env, rowvar=False, ddof=0) for env in data]
    n_obs = [len(env) for env in data]  # number of observations/environment
    return full(B, omegas, sample_covariances, n_obs)


def local_raw(j, b, omegas, data):
    """
    Compute the conditional likelihood of the marginal observations of
    a variable given observations of its hypothetical parents, given
    the weights (b) for those parents and the variable's error term
    variances over each environment (omegas).

    Parameters
    ----------
    j : int
        the variable's index
    b : np.array
        p-sized array with the weights for the parents of j and 0s
        elswhere
    omegas : np.array
        1d array with the noise term variance of j for each
        environment
    data : list(np.array)
        the complete joint observations for each environment, where
         each element of the list is an np.array of size n_e x p,
         where p is the total number of variables (same size as b) and
         n_e is the number of observations from that environment.

    Returns
    -------
    log_likelihood : float
        the resulting log_likelihood

    """
    sample_covariances = [np.cov(env, rowvar=False, ddof=0) for env in data]
    n_obs = [len(env) for env in data]
    return local(j, b, omegas, sample_covariances, n_obs)

# def local(j,b,omegas,data):
#     """New with manually computed scatter matrix"""
#     data = data[0]
#     n = len(data)
#     normalized = np.hstack([data, np.ones((n,1))])
#     normalized -= normalized.mean(axis=0)
#     cov = normalized.T @ normalized / n
#     pa = np.where(b != 0)[0]
#     if len(pa) > 0:
#         sub_cov = cov[:,pa][pa,:]
#         b = cov[j,pa]
#         sigma = cov[j,j] - b @ np.linalg.solve(sub_cov, b)
#         #likelihood = -n * np.log(2*np.pi) - n * np.log(sigma) - 1 / sigma
#     else:
#         sigma = cov[j,j]
#     n = len(data)
#     likelihood = -0.5*n*( 1 + np.log(sigma/n))
#     return likelihood

# def local(j,b,omegas,data):
#     """From raw data"""
#     data = data[0]
#     n,p = data.shape
#     data = np.hstack([data, np.ones((n,1))])
#     pa = list(np.where(b != 0)[0]) + [p]
#     Y = data[:,j]
#     X = np.atleast_2d(data[:,pa])
#     # QR-decomposition
#     # Q = np.linalg.qr(np.atleast_2d(data[:,pa]), 'complete')
#     # sigma = np.sum(Y**2) - np.sum((Y @ Q) ** 2)
#     # using solve
#     b = np.linalg.lstsq(X, Y, rcond=None)[0]
#     sigma = np.sum((Y - X @ b)**2)
#     likelihood = -0.5*n*( 1 + np.log(sigma/n))
#     return likelihood



# Copyright 2022 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import ges
from gnies.scores import GnIESScore as Score
import gnies.utils as utils

# TODO: Docstrings for known_targets
# TODO:
# - Made condition for improved score > instead of >=
# - Removed centered as argument _inner_procedure since this is defined in the score that is passed to it; add

# --------------------------------------------------------------------
# Public API


def fit(
    data,
    lmbda=None,
    known_targets=set(),
    approach="greedy",
    # Parameters used for greedy approach
    I0=set(),
    phases=["forward", "backward"],
    # Parameters used for rank approach
    direction="forward",
    center=True,
    # Parameters used by inner-procedure (modified GES)
    ges_iterate=True,
    ges_phases=["forward", "backward", "turning"],
    debug=0,
):
    """Runs the GnIES algorithm on the given data, producing an estimate
    of the I-equivalence class and the intervention targets.

    Parameters
    ----------
    data : list of numpy.ndarray
        A list with the samples from the different environments, where
        each sample is an array with columns corresponding to
        variables and rows to observations.
    lmbda : float, default=None
        The penalization parameter for the penalized-likelihood
        score. If `None`, the BIC penalization is chosen, that is,
        `0.5 * log(N)` where `N` is the total number of observations,
        pooled across environments.
    known_targets : set, default=set()
        The set of known intervention targets, which are always kept
        as part of the target estimate of the outer procedure.
    approach : {'greedy', 'rank'}, default='greedy'
        The approach used by the outer procedure of GnIES. 'greedy'
        means that intervention targets are greedily added/removed;
        'rank' means that an ordering is found by first fitting a
        model with `I={1,...,p}`, and targets are added/removed in
        this order.
    I0 : set, default=set()
        If the 'greedy' approach is selected, specifies the initial
        set of intervention targets, to which targets are
        added/removed.
    phases : [{'forward', 'backward'}*], default=['forward', 'backward']
        If the 'greedy' approach is selected, specifies which phases
        of the outer procedure are run.
    direction : {'forward', 'backward'}, default='forward'
        If the 'rank' approach is selected, specifies whether we add or
        remove variables: if 'forward', we start with an empty
        intervention set and add targets according to the found
        ordering; if 'backward', we start with the full set and remove
        targets instead.
    center : bool, default=True
        Whether GnIES considers interventions only on the noise-term
        variance (`center=True`, i.e., data is centered before
        computing the score) or on both the mean and variance
        (`center=False`).
    ges_iterate : bool, default=False
        Indicates whether the phases of the inner procedure (modified
        GES) should be iterated more than once.
    ges_phases : [{'forward', 'backward', 'turning'}*], optional
        Which phases of the inner procedure (modified GES) are run,
        and in which order. Defaults to `['forward', 'backward',
        'turning']`.
    debug : int, default=0
        If larger than 0, debug are traces printed. Higher values
        correspond to increased verbosity.

    Returns
    -------
    score : float
        The penalized likelihood score of the GnIES estimate.
    icpdag : numpy.ndarray
        The I-CPDAG representing the estimated I-equivalence class,
        where `icpdag[i,j] != 0` implies the edge `i -> j` and
        `icpdag[i,j] != 0 & icpdag[j,i] != 0` implies the edge `i -
        j`.
    I : set of ints
        The estimate of the intervention targets.

    Raises
    ------
    ValueError:
        If an invalid value is selected for the `approach` parameter.

    """
    if approach == "greedy":
        return fit_greedy(
            data,
            lmbda,
            known_targets,
            I0,
            phases,
            center,
            ges_iterate,
            ges_phases,
            debug,
        )
    elif approach == "rank":
        return fit_rank(
            data,
            lmbda,
            known_targets,
            direction,
            center,
            ges_iterate,
            ges_phases,
            debug,
        )
    else:
        raise ValueError('Invalid value "%s" for parameter "approach"' % approach)


def fit_greedy(
    data,
    lmbda=None,
    known_targets=set(),
    I0=set(),
    phases=["forward", "backward"],
    center=True,
    ges_iterate=True,
    ges_phases=["forward", "backward", "turning"],
    debug=0,
):
    """Run the outer procedure of GnIES by greedily adding/removing
    variables until the score does not improve.

    Parameters
    ----------
    data : list of numpy.ndarray
        A list with the samples from each environment, where each
        sample is an array with columns corresponding to variables and
        rows to observations.
    lmbda : float, default=None
        The penalization parameter for the penalized-likelihood
        score. If `None`, the BIC penalization is chosen, that is,
        `0.5 * log(N)` where `N` is the total number of observations,
        pooled across environments.
    known_targets : set, default=set()
        The set of known intervention targets, which are always kept
        as part of the target estimate of the outer procedure.
    I0 : set, default=set()
        Together with `known_targets`, this makes initial set of
        intervention targets, to which targets are added/removed.
    phases : [{'forward', 'backward'}*], default=['forward', 'backward']
        Specifies which phases of the outer procedure are run.
    center : bool, default=True
        Whether GnIES considers interventions only on the noise-term
        variance (`center=True`, i.e., data is centered before
        computing the score) or on both the mean and variance
        (`center=False`).
    ges_phases : [{'forward', 'backward', 'turning'}*], optional
        Which phases of the inner procedure (modified GES) are run,
        and in which order. Defaults to `['forward', 'backward',
        'turning']`.
    ges_iterate : bool, default=False
        Indicates whether the phases of the inner procedure (modified
        GES) should be iterated more than once.
    debug : int, default=0
        If larger than 0, debug are traces printed. Higher values
        correspond to increased verbosity.

    Returns
    -------
    score : float
        The penalized likelihood score of the GnIES estimate.
    icpdag : numpy.ndarray
        The I-CPDAG representing the estimated I-equivalence class,
        where `icpdag[i,j] != 0` implies the edge `i -> j` and
        `icpdag[i,j] != 0 & icpdag[j,i] != 0` implies the edge `i -
        j`.
    I : set of ints
        The estimate of the intervention targets.

    """

    print("Running GnIES with greedy phases %s" % phases) if debug else None
    # Inner procedure parameters
    params = {
        "phases": ges_phases,
        "iterate": ges_iterate,
        "debug": 2 if debug > 1 else 0,
    }

    # Iteration 0: initial set + known targets
    current_I = I0 | known_targets
    score_class = Score(data, current_I, lmbda=lmbda, centered=center)
    current_estimate, current_score = _inner_procedure(score_class, current_I, **params)

    # Iterate
    p = score_class.p
    full_I = set(range(p))
    phase = "forward"
    for phase in phases:
        print("  GnIES %s phase" % phase) if debug else None
        while True:
            assert known_targets <= current_I
            print(
                "    Current I=%s (score = %0.2f)" % (current_I, current_score)
            ) if debug else None
            scores = []
            next_Is = (
                full_I - current_I if phase == "forward" else current_I - known_targets
            )
            # If no more variables can be added/removed, end this phase
            if len(next_Is) == 0:
                break
            for i in next_Is:
                new_I = current_I | {i} if phase == "forward" else current_I - {i}
                score_class.set_I(new_I)
                estimate, score = _inner_procedure(score_class, new_I, **params)
                print("      Scored I=%s : %0.2f" % (new_I, score)) if debug else None
                scores.append((score, new_I, estimate))
            # Pick the maximally scoring addition/removal
            new_score, new_I, new_estimate = max(scores)
            # If the score was improved, repeat the greedy step
            if new_score > current_score:
                current_score, current_I, current_estimate = (
                    new_score,
                    new_I,
                    new_estimate,
                )
            # Otherwise, halt
            else:
                print("    Score was not improved.") if debug else None
                break
    return current_score, current_estimate, current_I


def fit_rank(
    data,
    lmbda=None,
    known_targets=set(),
    direction="forward",
    center=True,
    ges_iterate=True,
    ges_phases=["forward", "backward", "turning"],
    debug=0,
):
    """Run the outer procedure of GnIES; instead of greedily
    adding/removing intervention targets, use the ordering implied by
    the variance of the noise-term-variance estimates of each
    variable. The ordering is found by first fitting a model allowing
    interventions on all targets.

    Parameters
    ----------
    data : list of numpy.ndarray
        A list with the samples from each environment, where each
        sample is an array with columns corresponding to variables and
        rows to observations.
    lmbda : float, default=None
        The penalization parameter for the penalized-likelihood
        score. If `None`, the BIC penalization is chosen, that is,
        `0.5 * log(N)` where `N` is the total number of observations,
        pooled across environments.
    known_targets : set, default=set()
        The set of known intervention targets, which are always kept
        as part of the target estimate of the outer procedure.
    direction : {'forward', 'backward'}, default='forward'
        If 'forward', we start with an empty intervention set and add
        targets according to the found ordering. If 'backward', we
        start with the full set and remove targets.
    center : bool, default=True
        Whether GnIES considers interventions only on the noise-term
        variance (`center=True`, i.e., data is centered before
        computing the score) or on both the mean and variance
        (`center=False`).
    ges_phases : [{'forward', 'backward', 'turning'}*], optional
        Which phases of the inner procedure (modified GES) are run,
        and in which order. Defaults to `['forward', 'backward',
        'turning']`.
    ges_iterate : bool, default=False
        Indicates whether the phases of the inner procedure (modified
        GES) should be iterated more than once.
    debug : int, default=0
        If larger than 0, debug are traces printed. Higher values
        correspond to increased verbosity.

    Returns
    -------
    score : float
        The penalized likelihood score of the GnIES estimate.
    icpdag : numpy.ndarray
        The I-CPDAG representing the estimated I-equivalence class,
        where `icpdag[i,j] != 0` implies the edge `i -> j` and
        `icpdag[i,j] != 0 & icpdag[j,i] != 0` implies the edge `i -
        j`.
    I : set of ints
        The estimate of the intervention targets.

    """

    print("Running GnIES with %s-rank approach" % direction) if debug else None

    # Inner procedure parameters
    params = {
        "phases": ges_phases,
        "iterate": ges_iterate,
        "debug": 2 if debug > 1 else 0,
    }

    # First fit with full I to obtain an ordering based on the variance of the
    # noise-term variance estimates of each variable
    p = data[0].shape[1]
    e = len(data)
    full_I = set(range(p))
    score_class = Score(data, full_I, lmbda=lmbda, centered=center)
    current_estimate, current_score = _inner_procedure(score_class, full_I, **params)
    assert utils.is_dag(current_estimate)
    _, omegas, _ = score_class._mle_full(current_estimate)
    variances = np.var(omegas, axis=0)
    order = list(np.argsort(variances))
    # Setup for the greedy outer procedure
    if direction == "forward":
        current_I = known_targets
        score_class = Score(data, current_I, lmbda=lmbda)
        current_estimate, current_score = _inner_procedure(
            score_class, current_I, **params
        )
        verb = "Adding"
        order.reverse()
    elif direction == "backward":
        current_I = full_I
        verb = "Pruning"
    else:
        raise ValueError('Invalid value "%s" for field "order"' % order)

    # Add/remove intervention targets according to the obtained
    # ordering, until the score does not improve
    print("  %s intervention targets in order %s" % (verb, order)) if debug else None
    for i in order:
        assert known_targets <= current_I
        if i in known_targets:
            continue
        print(
            "    Current I=%s (score = %0.2f)" % (current_I, current_score)
        ) if debug else None
        next_I = current_I | {i} if direction == "forward" else current_I - {i}
        score_class.set_I(next_I)
        next_estimate, next_score = _inner_procedure(score_class, next_I, **params)
        if next_score > current_score:
            current_score, current_estimate, current_I = (
                next_score,
                next_estimate,
                next_I,
            )
        else:
            break
    # Return highest scoring estimate
    return current_score, current_estimate, current_I


# --------------------------------------------------------------------
# Auxiliary functions


def _inner_procedure(
    score_class,
    I,
    lmbda=None,
    phases=["forward", "backward", "turning"],
    iterate=True,
    covariances=None,
    previous_score=None,
    debug=0,
):
    """Run the inner procedure of GnIES, i.e. GES with a modified score
    and completion algorithm.

    Parameters
    ----------
    data : list of numpy.ndarray
        A list with the samples from each environment, where each
        sample is an array with columns corresponding to variables and
        rows to observations.
    I : set of ints
        The intervention targets used for the inner procedure.
    lmbda : float, default=None
        The penalization parameter for the penalized-likelihood
        score. If `None`, the BIC penalization is chosen, that is,
        `0.5 * log(N)` where `N` is the total number of observations,
        pooled across environments.
    phases : [{'forward', 'backward', 'turning'}*], optional
        Which phases of the inner procedure (modified GES) are run,
        and in which order. Defaults to `['forward', 'backward',
        'turning']`.
    iterate : bool, default=False
        Indicates whether the phases of the inner procedure (modified
        GES) should be iterated more than once.
    covariances : numpy.ndarray, default=None
        A `e x p x p` array where `e` is the number of environmnets
        and `p` the number of variables. Specifies the scatter
        matrices used to compute the score; if `None` (default), the
        scatter matrices are the sample covariance matrices.
    debug : int, default=0
        If larger than 0, debug are traces printed. Higher values
        correspond to increased verbosity.

    Returns
    -------

    """
    # Construct completion algorithm
    assert score_class.I == I

    def completion_algorithm(PDAG):
        return utils.pdag_to_icpdag(PDAG, I)

    # Run inner procedure
    estimate, score = ges.fit(
        score_class, completion_algorithm, phases=phases, iterate=iterate, debug=debug
    )

    return estimate, score
